# Imports
import pylab as pl
import sciris as sc
import scirisweb as sw
from datetime import datetime as dt

runserver = True # Choose to run in the frontend or backend

# Convert dates to years
def convertdate(datestr, fmt):
    date = dt.strptime(datestr, fmt)
    yearnum = date.year + (date.month-1)/12. + (date.day-1)/365.
    return yearnum

# Get the data
def loaddata():
    print('Loading data...')
    dataurl = 'https://raw.githubusercontent.com/rstudio/shiny-examples/master/120-goog-index/data/trend_data.csv'
    rawdata = sc.wget(dataurl).splitlines()
    data = []
    for r,rawline in enumerate(rawdata):
        line = rawline.split(',')
        if r==0: # Read header
            cols = line
        else: # Read data
            tag = line[0]
            yearnum = convertdate(line[1], '%Y-%m-%dT%I:%M:%fZ')
            value = float(line[2]) if r>0 else line[2]
            data.append([tag, yearnum, value])
    df = sc.dataframe(cols=cols, data=data)
    return df

# Create the app
app = sw.ScirisApp(__name__, name="PyShiny")
df = loaddata()

# Define the RPCs
@app.register_RPC()
def getoptions(tojson=True):
    options = sc.odict([
        ('Advertising',    'advert'),
        ('Education',      'educat'),
        ('Small business', 'smallbiz'),
        ('Travel',         'travel'),
        ('Unemployment',   'unempl'),
        ])
    if tojson:
        output = sc.sanitizejson(options.keys(), tostring=False)
    else:
        output = options
    return output

@app.register_RPC()
def plotdata(trendselection=None, startdate='2000-01-01', enddate='2018-01-01', trendline=False):
    
    # Handle inputs
    startyear = convertdate(startdate, '%Y-%m-%d')
    endyear   = convertdate(enddate,   '%Y-%m-%d')
    trendoptions = getoptions(tojson=False)
    if trendselection is None: trendselection  = trendoptions.keys()[0]
    datatype = trendoptions[trendselection]

    # Make graph
    fig = pl.figure()
    fig.add_subplot(111)
    thesedata = df.findrows(key=datatype, col='type')
    years = thesedata['date']
    vals = thesedata['close']
    validinds = sc.findinds(pl.logical_and(years>=startyear, years<=endyear))
    x = years[validinds]
    y = vals[validinds]
    pl.plot(x, y)
    pl.xlabel('Date')
    pl.ylabel('Trend index')
    
    # Add optional trendline
    if trendline:
        newy = sc.smoothinterp(x, x, y, smoothness=200)
        pl.plot(x, newy, lw=3)
    
    # Convert to FE
    graphjson = sw.mpld3ify(fig, jsonify=False)  # Convert to dict
    return graphjson  # Return the JSON representation of the Matplotlib figure

# Run the server
if __name__ == "__main__" and runserver:
    app.run()
else:
    plotdata()


"""
datastore.py -- code related to the Sciris database.
    
Last update: 2019sep30
"""


#################################################################
### Imports and global variables
#################################################################

# Imports
import os
import six
import atexit
import tempfile
import traceback
import shutil
import fnmatch
import redis
import sqlalchemy
import sciris as sc
from .sw_users import User
from .sw_tasks import Task


# Global variables
default_settingskey = '!DataStoreSettings'      # Key for holding DataStore settings
default_separator   = '::'                     # Define the separator between a key type and uid
max_key_length      = 255

#################################################################
### Classes
#################################################################

__all__ = ['Blob', 'DataStoreSettings', 'make_datastore', 'DataDir', 'copy_datastore']


class PickleError(Exception):
    """ This error gets raised if a blob in the datastore could not be unpickled """
    pass


class Blob(sc.prettyobj):
    ''' Wrapper for any Python object we want to store in the DataStore. '''
    
    def __init__(self, obj=None, key=None, objtype=None, uid=None, force=True):
        # Handle input arguments
        if uid is None: 
            if force:
                uid = sc.uuid()
            else:
                errormsg = 'DataStore: Not creating a new Blob UUID since force is set to False: key=%s, objtype=%s, uid=%s, obj=%s' % (key, objtype, uid, obj)
                raise Exception(errormsg)
        if not key: key = '%s%s%s' % (objtype, default_separator, uid)
        
        # Set attributes
        self.key      = key
        self.objtype  = objtype
        self.uid      = uid
        self.created  = sc.now()
        self.modified = [self.created]
        self.obj      = obj
        return None
    
    def update(self):
        ''' When the object is updated, append the current time to the modified list '''
        now = sc.now()
        self.modified.append(now)
        return now
        
    def save(self, obj):
        ''' Save new object to the Blob '''
        self.obj = obj
        self.update()
        return None
    
    def load(self):
        ''' Load data from the Blob '''
        output = self.obj
        return output


def make_datastore(url=None, *args, **kwargs):
    """
    Make a datastore -- interface for the DataStore classes.

    :param url: URL that identifies a database. It can be a Redis URL, a file location, or a URL supported by SQLALchemy
        - Redis: 'redis://127.0.0.1:6379/8' [default]
        - Filesystem 'file:///home/username/storage' or 'file://./storage' for a relative path
        - SQLALchemy: some examples of URLs for various backends
            - 'sqlite:///storage.db'
            - 'mssql+pyodbc:///?odbc_connect=DRIVER%3D%7BODBC+Driver+13+for+SQL+Server%7D%3BSERVER%3D127.0.0.1%3BDATABASE%3Dtestdb%3BUID%3Dusername%3BPWD%3Dpassword%3B'-
            - 'postgresql+psycopg2://username:password@localhost:5432/testdb'
            - 'mysql+pymysql://username:password@localhost:3306/testdb?charset=utf8mb4&binary_prefix=true' (note the binary_prefix and charset components in the URL)
    :param args: Extra arguments to `DataStore`
    :param kwargs: Extra keyword arguments to `DataStore`
    :return: A `DataStore` instance of the appropriate derived type e.g. `RedisDataStore` for a Redis URL
    
    Examples:
        ds = sw.make_datastore() # Create a Redis database using defaults
        ds = sw.make_datastore('redis') # Likewise
        ds = sw.make_datastore('redis://127.0.0.1:6379/8') # Set up with a specific URL
        ds = sw.make_datastore(url='redis://127.0.0.1:6379/8', )

    """

    if url is None or url.startswith('redis'):
        if url == 'redis': url = None # Reset if not an actual URL
        return RedisDataStore(url, *args, **kwargs)
    elif url.startswith('file'): # Reset if not an actual URL
        if url == 'file': url = None
        return FileDataStore(url, *args, **kwargs)
    else:
        return SQLDataStore(url, *args, **kwargs)


def copy_datastore(src, dst):
    """
    Copy datastore so that the destination datastore is an replica of the source datastore

    'Hidden' keys starting with '_' will not be copied. This is important because keys like
    ``_kombu*`` created by Redis do not have a string type and thus cannot be moved between
    datastore backends.

    :param src: Datastore source URL
    :param dst: Datastore destination URL
    :return: None
    """

    src_ds = make_datastore(src)
    dst_ds = make_datastore(dst)
    dst_ds.flushdb()
    dst_ds = make_datastore(dst)
    keys = list(src_ds.keys())
    for i, key in enumerate(keys):
        print("Key %s (%d of %d)" % (key, i, len(keys)))
        if key.startswith('_'):
            print("Skipping %s" % (key))
            continue
        value = src_ds._get(key)
        dst_ds._set(key, value)


class DataStoreSettings(sc.prettyobj):
    ''' Global settings for the DataStore '''
    
    def __init__(self, settings=None, tempfolder=None, separator=None):
        ''' Initialize with highest priority given to the inputs, then the stored settings, then the defaults '''
        
        # 1. Arguments
        self.tempfolder = tempfolder
        self.separator  = separator
        
        # 2. Existing settings
        if not settings:
            self.is_new    = True
            old_tempfolder = None
            old_separator  = None
        else:
            self.is_new    = False
            old_tempfolder = settings.tempfolder
            old_separator  = settings.separator
        
        # 3. Defaults
        def_tempfolder = tempfile.mkdtemp()
        def_separator  = default_separator
        
        # Iterate in order 
        tempfolder_list = [old_tempfolder, def_tempfolder]
        separator_list  = [old_separator,  def_separator]
        for folder in tempfolder_list:
            if not self.tempfolder:
                self.tempfolder = folder
        for sep in separator_list:
            if not self.separator:
                self.separator = sep
        
        return None



class BaseDataStore(sc.prettyobj):
    """
    Base DataStore functionality

    This base class implements primary DataStore functionality independent of the specific
    storage backend being used. Derived classes implement the private methods `_set`, `_get`
    etc. as required for their specific storage mechanisms. This class implements the logical operations
    for constructing keys, storing users and tasks etc.

    This class should not be instantiated directly - instead, construct a datastore either using
    the `datastore()` function, or by instantiating one of the backend-specific datastores e.g.
    `RedisDataStore`.

    """

    def __init__(self, tempfolder=None, separator=None, settingskey=None, verbose=True):
        self.tempfolder = None # Populated by self.settings()
        self.separator  = None # Populated by self.settings()
        self.is_new     = None # Populated by self.settings()
        self.verbose    = verbose
        self.settings(settingskey=settingskey, tempfolder=tempfolder, separator=separator) # Set or get the settings
        if self.verbose: print(self)
        return None

    ### DATASTORE BACKEND-SPECIFIC METHODS, THAT NEED TO BE DEFINED IN DERIVED CLASSES FOR SPECIFIC STORAGE MECHANISMS E.G. REDIS

    def __repr__(self):
        pass


    def _set(self, key , objstr):
        """
        Store string content under key

        This method is assumed to succeed unless an error is raised. If the key is
        already present, the specification is that this method will overwrite it.
        Logic for dealing with the case where the key already exists
        should be implemented in `DataStore.set()` so that it is common to all backends.

        :param key: Database key to store the object under
        :param objstr: Binary string with the string representation of the object
        :return: `None` if operation was successful
        :raises: `Exception` if operation failed
        """
        pass


    def _get(self, key):
        """
        Return blob content as string

        Derived classes implementing this method should return `None` if the key is not present.
        The logic of what happens next is implemented in `DataStore.get()` and is common
        to all storage backends

        :param key: Database key the object is stored under
        :return: Binary string corresponding to the object. `None` if the key was not present

        """
        pass


    def _delete(self, key):
        """
        Remove entry from database

        This function is assumed to succeed unless an error is raised. If the key is not
        present, this function is considered to have succeeded and an error should
        not be raised.

        :param key: Database key the object is stored under
        :return: `None` if delete succeeded
        :raises: `Exception` if delete was not successful

        """
        pass


    def _flushdb(self):
        """
        Clear all content from the datastore

        :return: `None` if flush succeeded
        :raises: `Exception` if flush was not successful
        """
        pass


    def _keys(self):
        """
        Return all keys in datastore

        :return: List of keys
        """
        pass


    ### STANDARD DATASTORE FUNCTIONALITY

    def set(self, key=None, obj=None, objtype=None, uid=None):
        """
        Store item in datastore

        Operation is considered to succeed if this method does not raise an error

        TODO - Could add checks like checking if it's a Blob instance (but settings are not stored as blobs...)

        :param key:
        :param obj: A Blob instance
        :param objtype:
        :param uid:
        :return:

        """

        key = self.getkey(key=key, objtype=objtype, uid=uid, obj=obj)
        objstr = sc.dumpstr(obj)
        self._set(key, objstr)
        return None


    def get(self, key=None, obj=None, objtype=None, uid=None, notnone=False, die=False):
        """
        Retrieve item from datastore

        If nonnone is False, then if the key is missing, `None` will be returned.
        If the key does exist, then an sw.Blob instance will be returned

        :param key:
        :param obj: A Blob instance
        :param objtype:
        :param uid:
        :param notnone:
        :param die:
        :return:

        :raises: KeyError if key is not present. PickleError if the blob was present but could not be unpickled
        """

        key = self.getkey(key=key, objtype=objtype, uid=uid, obj=obj)

        objstr = self._get(key)

        if objstr is None and notnone:
            errormsg = 'Datastore key "%s" not found (obj=%s, objtype=%s, uid=%s)' % (key, obj, objtype, uid)
            raise KeyError(errormsg)
        elif objstr is None:
            return None

        try:
            output = sc.loadstr(objstr, die=die)
        except:
            output = None
            errormsg = 'Datastore error: unpickling failed:\n%s' % traceback.format_exc()  # Grab the trackback stack
            if die:
                raise PickleError(errormsg)
            else:
                print(errormsg)
        return output


    def delete(self, key=None, obj=None, objtype=None, uid=None, die=None):
        """
        Remove item from datastore

        :param key:
        :param obj:
        :param objtype:
        :param uid:
        :param die:
        :return:
        """
        key = self.getkey(key=key, objtype=objtype, uid=uid, obj=obj)
        self._delete(key)
        if self.verbose: print('DataStore: deleted key %s' % key)
        return None


    def exists(self, key):
        """
        Return True if key exists in the datastore

        Can overload this if the specific datastore implements
        a faster method for checking

        :param key:
        :return: True if the key exists, False otherwise

        """
        objstr = self._get(key)
        return objstr is not None


    def flushdb(self):
        self._flushdb()
        if self.verbose: print('DataStore flushed.')
        return None


    def keys(self, pattern=None):
        """
        Return list of keys, optionally filtered

        :param pattern: Regular expression, key will be retained if a search for this expression returns a result
        :return: List of keys
        """
        keys = self._keys()
        if pattern is not None:
            keys = [x for x in keys if fnmatch.fnmatch(x, pattern)]  # Use fnmatch rather than re to mirror Redis's built-in behaviour
        return keys


    def settings(self, settingskey=None, tempfolder=None, separator=None, die=False):
        ''' Handle the DataStore settings '''
        if not settingskey: settingskey = default_settingskey
        try:
            origsettings = self.get(settingskey)
        except Exception as E:
            origsettings = None
            errormsg = 'Datastore: warning, could not load settings, using defaults: %s' % str(E)
            if die: raise Exception(errormsg)
            else:   print(errormsg)
        settings = DataStoreSettings(settings=origsettings, tempfolder=tempfolder, separator=separator)
        self.tempfolder = settings.tempfolder
        self.separator  = settings.separator
        self.is_new     = settings.is_new
        self.set(settingskey, settings) # Save back to the database
        
        # Handle the temporary folder
        try:
            os.makedirs(self.tempfolder)
            atexit.register(self._rmtempfolder) # Only register this if we've just created the temp folder
        except FileExistsError:
            pass

        return settings

    
    def _rmtempfolder(self):
        ''' Remove the temporary folder that was created '''
        if os.path.exists(self.tempfolder):
            if self.verbose: print('Removing up temporary folder at %s' % self.tempfolder)
            shutil.rmtree(self.tempfolder)
        return None
    
    
    def makekey(self, objtype, uid):
        ''' Create a key from an object type and UID '''
        if objtype: key = '%s%s%s' % (objtype, self.separator, uid) # Construct a key with an object type and separator
        else:       key = '%s'     % uid                            # ...or, just return the UID
        return key
        
    
    def getkey(self, key=None, objtype=None, uid=None, obj=None, fulloutput=None, forcetype=None):
        '''
        Get a valid database key, either from a given key (do nothing), or else from
        a supplied objtype and uid, or else read them from the object supplied. The
        idea is for this method to be as forgiving as possible for different possible
        combinations of inputs.
        '''
        # Handle optional input arguments
        if fulloutput is None: fulloutput = False
        if forcetype  is None: forcetype  = True
        
        # Handle different sources for things
        props = ['key', 'objtype', 'uid']
        args    = {'key':key,  'objtype':objtype, 'uid':uid}  # These are what have been supplied by the user
        fromobj = {'key':None, 'objtype':None,    'uid':None} # This is from the object
        final   = {'key':None, 'objtype':None,    'uid':None} # These will eventually be the output values -- copy of args
        
        # Look for missing properties from the object
        if obj:
            if hasattr(obj, 'key'):     fromobj['key']     = obj.key
            if hasattr(obj, 'objtype'): fromobj['objtype'] = obj.objtype
            if hasattr(obj, 'uid'):     fromobj['uid']     = obj.uid
        
        # Populate all non-None entries from the input arguments
        for p in props:
            if args[p]:
                final[p] = sc.flexstr(args[p]) # Convert to string since you don't know what crazy thing might be passed (using flexstr since str can't handle bytes)

        # Populate what we can from the object, if it hasn't already been populated
        for p in props:
            if fromobj[p] and not final[p]:
                final[p] = fromobj[p]
        
        # If the key is supplied but other things aren't, try to create them now
        if final['key'] and (not final['objtype'] or not final['uid']):
            splitkey = final['key'].split(self.separator)
            if len(splitkey)==2: # Check that the key split properly
                if not final['objtype']: final['objtype'] = splitkey[0]
                if not final['uid']:     final['uid']     = splitkey[1]
        
        # If we're still stuck, try making a new uid
        if not final['key'] and not final['uid']:
            final['uid'] = str(sc.uuid())
        
        # If everything is supplied except the key, create it
        if not final['key']:
            if final['objtype'] and final['uid']: # Construct a key from the object type and UID
                final['key'] = self.makekey(objtype=final['objtype'], uid=final['uid'])
            elif not final['objtype'] and final['uid']: # Otherwise, just use the UID
                final['key'] = final['uid']
            
        # Check that it's found, and if not, treat the key as a UID and try again
        keyexists = self.exists(final['key']) # Check to see whether a match has been found
        if not keyexists: # If not, treat the key as a UID instead
            newkey = self.makekey(objtype=final['objtype'], uid=final['key'])
            newkeyexists = self.exists(newkey) # Check to see whether a match has been found
            if newkeyexists:
                final['key'] = newkey
        
        # Finally, force the type if requested
        if forcetype and final['objtype']:
            splitkey = final['key'].split(self.separator, 1)
            if splitkey[0] != final['objtype']:
                final['key'] = self.makekey(objtype=final['objtype'], uid=final['key'])

        if len(final['key']) > max_key_length:
            raise Exception('Key is too long')

        # Return what we need to return
        if fulloutput: return final['key'], final['objtype'], final['uid']
        else:          return final['key']


    def items(self, pattern=None):
        ''' Return all found items in an odict '''
        output = sc.odict()
        keys = self.keys(pattern=pattern)
        for key in keys:
            output[key] = self.get(key)
        return output
    
    
    def _checktype(self, key, obj, objtype):
        if   objtype == 'Blob': objclass = Blob
        elif objtype == 'User': objclass = User
        elif objtype == 'Task': objclass = Task
        else:
            errormsg = 'Unrecognized type "%s": must be Blob, User, or Task'
            raise Exception(errormsg)
        
        if obj is None:
            errormsg = 'Cannot load %s as a %s: key not found' % (key, objtype)
            raise Exception(errormsg)
        
        if not isinstance(obj, objclass):
            errormsg = 'Cannot load %s as a %s since it is %s' % (key, objtype, type(obj))
            raise Exception(errormsg)
        
        return None
        
    
    def saveblob(self, obj, key=None, objtype=None, uid=None, overwrite=None, forcetype=None, die=None):
        '''
        Add a new or update existing Blob in the datastore, returns key. If key is None,
        constructs a key from the Blob (objtype:uid); otherwise, updates the Blob with the 
        provided key.
        '''
        # Set default arguments
        if overwrite is None: overwrite = True
        if die       is None: die       = True
        
        key, objtype, uid = self.getkey(key=key, objtype=objtype, uid=uid, obj=obj, fulloutput=True, forcetype=forcetype)
        blob = self.get(key)
        if blob:
            self._checktype(key, blob, 'Blob')
            if overwrite:
                blob.save(obj)
            else:
                errormsg = 'DataStore: Blob %s already exists and overwrite is set to False' % key
                if die: raise Exception(errormsg)
                else:   print(errormsg)
        else:
            blob = Blob(key=key, objtype=objtype, uid=uid, obj=obj)
        self.set(key=key, obj=blob)
        if self.verbose: print('DataStore: Blob "%s" saved' % key)
        return key
    
    
    def loadblob(self, key=None, objtype=None, uid=None, forcetype=None, die=None):
        ''' Load a blob from the datastore '''
        if die is None: die = True
        key = self.getkey(key=key, objtype=objtype, uid=uid, forcetype=forcetype)
        blob = self.get(key)
        if die: self._checktype(key, blob, 'Blob')
        if isinstance(blob, Blob):
            obj = blob.load()
            if self.verbose: print('DataStore: Blob "%s" loaded' % key)
            return obj
        else:
            if self.verbose: print('DataStore: Blob "%s" not found' % key)
            return None
    
    
    def saveuser(self, user, overwrite=True, forcetype=None, die=None):
        '''
        Add a new or update existing User in Redis, returns key.
        '''
        if die is None: die = True
        key, objtype, username = self.getkey(objtype='user', uid=user.username, fulloutput=True, forcetype=forcetype)
        olduser = self.get(key)
        if olduser and not overwrite:
            errormsg = 'DataStore: User %s already exists, not overwriting' % key
            if die: raise Exception(errormsg)
            else:   print(errormsg)
        else:
            self._checktype(key, user, 'User')
            self.set(key=key, obj=user)
            if self.verbose:
                if olduser: print('DataStore: User "%s" updated' % key)
                else:       print('DataStore: User "%s" created' % key)
        return key
    
    
    def loaduser(self, username=None, key=None, forcetype=None, die=None):
        ''' Load a user from Redis '''
        if die is None: die = True
        key = self.getkey(key=key, objtype='user', uid=username, forcetype=forcetype)
        user = self.get(key)
        if die: self._checktype(key, user, 'User')
        if isinstance(user, User):
            if self.verbose: print('DataStore: User "%s" loaded' % key)
            return user
        else:
            if self.verbose: print('DataStore: User "%s" not found' % key)
            return None
        
    
    def savetask(self, task, key=None, uid=None, overwrite=None, forcetype=None):
        '''
        Add a new or update existing Task in Redis, returns key.
        '''
        if overwrite is None: overwrite = True
        key, objtype, uid = self.getkey(key=key, objtype='task', uid=uid, obj=task, fulloutput=True, forcetype=forcetype)
        oldtask = self.get(key)
        if oldtask and not overwrite:
            errormsg = 'DataStore: Task %s already exists' % key
            raise Exception(errormsg)
        else:
            self._checktype(key, task, 'Task')
            self.set(key=key, obj=task)
        if self.verbose: print('DataStore: Task "%s" saved' % key)
        return key
    
    
    def loadtask(self, key=None, uid=None, forcetype=None, die=None):
        ''' Load a user from Redis '''
        if die is None: die = False # Here, we won't always know whether the task exists
        key = self.getkey(key=key, objtype='task', uid=uid, forcetype=forcetype)
        task = self.get(key)
        if die: self._checktype(key, task, 'Task')
        if isinstance(task, Task):
            if self.verbose: print('DataStore: Task "%s" loaded' % key)
            return task
        else:
            if self.verbose: print('DataStore: Task "%s" not found' % key)
            return None



class RedisDataStore(BaseDataStore):
    """
    DataStore backed by Redis.
    
    :param redisargs: Arguments passed to the Redis constructor
    """

    def __init__(self, url=None, redisargs=None, *args, **kwargs):

        # Handle arguments to Redis (or lack thereof)
        if redisargs is None: 
            redisargs = {}
        
        # Handle the Redis URL
        default_url = 'redis://127.0.0.1:6379/'  # The default URL for the Redis database
        if not url:            url = default_url + '0' # e.g. sw.DataStore()
        elif sc.isnumber(url): url = default_url + '%i'%url # e.g. sw.DataStore(3)
        self.url = url
        self.redis = redis.StrictRedis.from_url(self.url, **redisargs)

        # Finish construction
        if six.PY2:
            super(RedisDataStore, self).__init__(*args, **kwargs)
        else:
            super().__init__(*args, **kwargs)
        return None
    

    ### DEFINE MANDATORY FUNCTIONS

    def __repr__(self):
        return '<RedisDataStore at %s with temp folder %s>' % (self.url, self.tempfolder)

    def _set(self, key, objstr):
        if six.PY3:
            key = key.encode()
        return self.redis.set(key, objstr)


    def _get(self, key):
        ''' Alias to redis.get() '''
        if six.PY3:
            key = key.encode()
        return self.redis.get(key)


    def _delete(self, key):
        self.redis.delete(key)
        return None


    def _flushdb(self):
        self.redis.flushdb()
        return None
        

    def _keys(self):
        keys = list(self.redis.keys())
        if six.PY3:
            keys = [x.decode() for x in keys]
        return keys


    ### OVERLOAD ADDITIONAL METHODS WITH REDIS BUILT-INS

    def keys(self, pattern=None):
        """
        Filter keys in redis to increase performance

        :param pattern:
        :return:
        """
        if pattern is None: pattern = '*'
        keys = list(self.redis.keys(pattern=pattern))
        if six.PY3:
            keys = [x.decode() for x in keys]
        return keys


    def exists(self, key):
        """
        Use Redis built-in exists function
        """
        output = self.redis.exists(key)
        return bool(output)



class SQLDataStore(BaseDataStore):
    """
    DataStore backed by SQLAlchemy/SQL
    """

    def __init__(self, url, sqlargs=None, *args, **kwargs):
        
        if sqlargs is None:
            sqlargs = {}

        if url is not None:
            self.url = url
        else:
            errormsg = 'To create an SQL DataStore, you must supply the URL'
            raise Exception(errormsg)

        # Define the internal class that is mapped to the SQL database
        from sqlalchemy.ext.declarative import declarative_base
        Base = declarative_base()

        class SQLBlob(Base):
            __tablename__ = 'datastore'
            key = sqlalchemy.Column('key', sqlalchemy.types.String(length=max_key_length), primary_key=True)
            content = sqlalchemy.Column('blob', sqlalchemy.types.LargeBinary)
        self.datatype = SQLBlob  # The class to use when interfacing with the database

        # Create the database
        self.engine = sqlalchemy.create_engine(self.url, **sqlargs)
        Base.metadata.create_all(self.engine)
        self.get_session = sqlalchemy.orm.session.sessionmaker(bind=self.engine)

        # Finish construction
        if six.PY2:
            super(SQLDataStore, self).__init__(*args, **kwargs)
        else:
            super().__init__(*args, **kwargs)
        return None


    ### DEFINE MANDATORY FUNCTIONS

    def __repr__(self):
        return '<SQLDataStore (%s) with temp folder %s>' % (self.url, self.tempfolder)


    def _set(self, key, objstr):
        session = self.get_session()
        obj = session.query(self.datatype).get(key)
        if obj is None:
            obj = self.datatype(key=key, content=objstr)
            session.add(obj)
        else:
            obj.content = objstr
        session.commit()
        session.close()
        return None


    def _get(self, key):
        session = self.get_session()
        obj = session.query(self.datatype).get(key)
        session.close()
        if obj is None:
            return None
        else:
            return obj.content


    def _delete(self, key):
        session = self.get_session()
        session.query(self.datatype).filter(self.datatype.key==key).delete()
        session.commit()
        session.close()
        return None


    def _flushdb(self):
        # Flush DB by dropping table - this allows the table schema to be changed
        # when the datastore is next instantiated
        self.datatype.__table__.drop(self.engine)
        self.engine.dispose()
        return None
    

    def _keys(self):
        session = self.get_session()
        keys = session.query(self.datatype.key).all() # Get all the keys
        session.close()
        return [x[0] for x in keys]



class FileDataStore(BaseDataStore):
    """
    DataStore backed by file-system storage

    WARNING - this backend may encounter errors if the files become locked by
    the OS due to another program using them (including another thread)
    """
    def __init__(self, url=None, suffix=None, prefix=None, dir=None, *args, **kwargs):
        
        if url is None: # It's not supplied, make a temporary folder
            self.path = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir) # Try to create a temporary directory
        else: # It's supplied, make sure it's in the right format
            self.path = os.path.abspath(url.replace('file://','')) + os.path.sep
            if not os.path.exists(self.path):
                os.makedirs(self.path)

        if six.PY2:
            super(FileDataStore, self).__init__(*args, **kwargs)
        else:
            super().__init__(*args, **kwargs)
        return None


    ### DEFINE MANDATORY FUNCTIONS

    def __repr__(self):
        return '<FileDataStore (%s) with temp folder %s>' % (self.path, self.tempfolder)


    def _set(self, key, objstr):
        with open(self.path + key,'wb') as f:
            f.write(objstr)
        return None


    def _get(self, key):
        if os.path.exists(self.path + key):
            with open(self.path + key,'rb') as f:
                objstr = f.read()
            return objstr
        else:
            return None


    def _delete(self, key):
        if os.path.exists(self.path + key):
            os.remove(self.path + key)
        return None


    def _flushdb(self):
        shutil.rmtree(self.path)
        os.mkdir(self.path)
        return None


    def _keys(self):
        keys = os.listdir(self.path)
        return keys


    ### OVERLOAD ADDITIONAL METHODS WITH FILE SYSTEM BUILT-INS

    def exists(self, key):
        return os.path.exists(self.path + key)


class DataDir(sc.prettyobj):
    ''' Alongside/instead of a DataStore, simply create a temporary folder to store essentials (e.g. uploaded files) '''
    
    def __init__(self, die=False, *args, **kwargs):
        try:
            if six.PY2: # Python 2
                self.tempfolder = tempfile.mkdtemp(*args, **kwargs) # If no datastore, try to create a temporary directory
                self.tempdir_obj = None
            else: # Python 3
                self.tempdir_obj = tempfile.TemporaryDirectory(*args, **kwargs) # If no datastore, try to create a temporary directory
                self.tempfolder = self.tempdir_obj.name # Save an alias to the directory name to match the DataStore object
                atexit.register(self.tempdir_obj.cleanup) # Only register this if we've just created the temp folder
        except:
            exception = traceback.format_exc() # Grab the trackback stack
            errormsg = 'Could not create temporary folder: %s' % exception
            self.tempfolder = errormsg # Store the error message in lieu of the folder name
            if die: raise Exception(errormsg)
            else:   print(errormsg) # Try to proceed if no datastore and the temporary directory can't be created
        return None


"""
tasks.py -- code related to Sciris task queue management
    
Last update: 2018sep23
"""

import traceback
from functools import wraps
from celery import Celery
from time import sleep
import sciris as sc
import scirisweb as sw
from . import sw_rpcs as rpcs


################################################################################
### Globals
################################################################################

__all__ = ['celery_instance'] # Others for internal use only

datastore = None
task_func_dict = {} # Dictionary to hold registered task functions to be callable from run_task().
RPC_dict = {} # Dictionary to hold all of the registered RPCs in this module.
RPC = rpcs.RPCwrapper(RPC_dict) # RPC registration decorator factory created using call to make_RPC().
celery_instance = None # Celery instance.



################################################################################
### Classes
################################################################################

__all__ += ['Task']


class Task(sc.prettyobj):
    '''
    A Sciris record for an asynchronous task.
    
    Attributes:
        task_id (str)         -- the ID / name for the task that typically is chosen by the client
        status (str)          -- the status of the task:
                                    'unknown': unknown status, for example, just initialized
                                    'error':   task failed with an actual error
        error_text (str)      -- string giving an idea of what error has transpired
        func_name (str)       -- string of the function name for what's called
        args (list)           -- list containing args for the function
        kwargs (dict)         -- dict containing kwargs for the function
        result_id (str)       -- string for the Redis ID of the AsyncResult
        queue_time (datetime) -- the time the task was queued for Celery
        start_time (datetime) -- the time the task was actually started
        stop_time (datetime)  -- the time the task completed
        pending_time (int)    -- the time the process has been waiting to be executed on the server in seconds        
        execution_time (int)  -- the time the process required to complete
    '''
    
    def  __init__(self, task_id):
        self.task_id        = task_id # Set the task ID (what the client typically knows as the task).
        self.uid            = task_id # Make it the same as the task ID...WARNING, fix
        self.status         = 'unknown' # Start the status at 'unknown'.
        self.error_msg      = None # Start the error_msg at None.
        self.error_text     = None # Start the error_text at None.
        self.func_name      = None # Start the func_name at None.
        self.args           = None # Start the args and kwargs at None.
        self.kwargs         = None
        self.result_id      = None # Start with no result_id.
        self.queue_time     = None # Start the queue, start, and stop times at None.
        self.start_time     = None
        self.stop_time      = None
        self.pending_time   = None # Start the pending and execution times at None.
        self.execution_time = None
        return None
    
    def show(self):
        print('-----------------------------')
        print('        Task ID: %s'   % self.task_id)
        print('         Status: %s'   % self.status)
        print('      Error msg: %s'   % self.error_msg)
        print('Error traceback: %s'   % self.error_text)
        print('  Function name: %s'   % self.func_name)
        print('  Function args: %s'   % self.args)
        print('Function kwargs: %s'   % self.kwargs)        
        print('      Result ID: %s'   % self.result_id)
        print('     Queue time: %s'   % self.queue_time)        
        print('     Start time: %s'   % self.start_time)
        print('      Stop time: %s'   % self.stop_time)
        print('   Pending time: %s s' % self.pending_time)        
        print(' Execution time: %s s' % self.execution_time)  
        print('-----------------------------')

    def jsonify(self):
        output = {'task':
                     {'UID':           self.uid,                    
                      'taskId':        self.task_id,
                      'status':        self.status,
                      'errorMsg':      self.error_msg,
                      'errorText':     self.error_text,
                      'funcName':      self.func_name,
                      'funcArgs':      self.args,
                      'funcKwargs':    self.kwargs,
                      'resultId':      self.result_id,
                      'queueTime':     self.queue_time,                
                      'startTime':     self.start_time,
                      'stopTime':      self.stop_time,
                      'pendingTime':   self.pending_time,
                      'executionTime': self.execution_time     
                      }
                  }
        return output

    def __setstate__(self, state):

        ### Migration for changing errorText to errorMsg+errorText
        if 'error_msg' not in state:
            state['error_msg'] = 'Error occured - please check console'
        self.__dict__ = state


################################################################################
### Functions
################################################################################

__all__ += ['get_datastore', 'make_celery', 'add_task_funcs', 'check_task', 'get_task_result', 'delete_task', 'taskwrapper']


def get_datastore(config=None):
    ''' Only if we have not already done so, create the DataStore object '''
    from . import sw_datastore as ds # This needs to be here to avoid a circular import
    try:
        datastore = sw.flaskapp.datastore
        assert datastore is not None
    except:
        if isinstance(config, dict): datastore_url = config['DATASTORE_URL']
        else:                        datastore_url = config.DATASTORE_URL
        datastore = ds.make_datastore(url=datastore_url)
    return datastore


# Function for creating the Celery instance, resetting the global and also 
# passing back the same result. for the benefit of callers in non-Sciris 
# modules.
def make_celery(config=None, verbose=True):
    global celery_instance
    global run_task_lock
    global datastore # So it's accessible in other functions

    run_task_lock = False
    celery_instance = Celery('tasks') # Define the Celery instance.
    if config is not None: # Configure Celery with config.py.
        celery_instance.config_from_object(config)
    datastore = get_datastore(config=config)

    # Define subfunctions available to the Celery instance
    
    def lock_run_task(task_id):
        # NOTE: We may want to change the lock mechanism to be tied to the task_id instead of using a global
        # run_task_lock, but only if we find out that there are conflicts with access to TaskRecord objects.
        global run_task_lock
        sleepduration = 5 # Define how lon to sleep before trying again
        if verbose: print('C>> Checking lock on run_task() for %s' % task_id)
        while run_task_lock: # Until there is no lock, sit and sleep.
            if verbose: print('C>> Detected lock on %s. Waiting...' % task_id)
            sleep(sleepduration)  # sleep before trying again
        if verbose: print('C>> No lock detected on %s, locking run_task()' % task_id)
        run_task_lock = True # Set the lock to keep other run_task() instances from co-occurring on the same Celery worker.
        return None
    
    
    def unlock_run_task(task_id):
        global run_task_lock
        if verbose: print('C>> Unlocking run_task() for %s' % task_id)
        run_task_lock = False # Remove the lock for this Celery worker.
        return None
    
    
    @celery_instance.task
    def run_task(task_id, func_name, args, kwargs):
        if kwargs is None: kwargs = {} # So **kwargs works below
        
        if verbose: print('C>> Starting run_task() for %s' % task_id)

        # Check if run_task() locked and wait until it isn't, then lock it for
        # other run_task() instances in this Celery worker.
        # NOTE: We may want to resurrect this, perhaps using the task_id as the lock, if we find out there are
        # conflicts with access to the TaskRecords.
        # lock_run_task(task_id)

        # Find a matching task record (if any) to the task_id.
        match_taskrec = datastore.loadtask(task_id)
        if match_taskrec is None:
            if verbose: print('C>> Failed to find task record for %s' % task_id)
            # unlock_run_task(task_id)  # uncomment if there are conflicts
            return { 'error': 'Could not access Task' }
    
        # Set the TaskRecord to indicate start of the task.
        match_taskrec.status = 'started'
        match_taskrec.start_time = sc.now()
        match_taskrec.pending_time = (match_taskrec.start_time - match_taskrec.queue_time).total_seconds()
            
        # Do the actual update of the TaskRecord.
        datastore.savetask(match_taskrec)
        if verbose: print('C>> Saved task for %s' % task_id)
        
        # Make the actual function call, inside a try block in case there is 
        # an exception thrown.
        # NOTE: This block is likely to run for several seconds or even 
        # minutes or hours, depending on the task.
        try:
            result = task_func_dict[func_name](*args, **kwargs)
            match_taskrec.status = 'completed'
            if verbose: print('C>> Successfully completed task %s! :)' % task_id)
        except Exception as e: # If there's an exception, grab the stack track and set the TaskRecord to have stopped on in error.
            error_text = traceback.format_exc()
            match_taskrec.status = 'error'
            match_taskrec.error_text = error_text
            match_taskrec.error_msg = str(e)
            result = error_text
            if verbose: print('C>> Failed task %s! :(' % task_id)
        
        # Set the TaskRecord to indicate end of the task.
        match_taskrec.stop_time = sc.now()
        match_taskrec.execution_time = (match_taskrec.stop_time - match_taskrec.start_time).total_seconds()
        
        # Do the actual update of the TaskRecord.  Do this in a try / except 
        # block because this step may fail.  For example, if a TaskRecord is 
        # deleted by the webapp, the update here will crash.
        try:
            datastore.savetask(match_taskrec)
        except Exception as e:  # If there's an exception, grab the stack track and set the TaskRecord to have stopped on in error.
            error_text = traceback.format_exc()
            match_taskrec.status = 'error'
            match_taskrec.error_text = error_text
            match_taskrec.error_msg = str(e)
            result = error_text
            if verbose: print('C>> Failed to save task %s! :(' % task_id)            
            
        if verbose: print('C>> End of run_task() for %s' % task_id)

        # Unlock run-task() for other run_task() instances running on the same
        # Celery worker.
        # unlock_run_task(task_id)  # uncomment if there are conflicts

        # Return the result.
        return result 
    
    # The launch-task() RPC is the only one included here because it is the 
    # only one that makes a direct call to run_task().  Any other RPCs that 
    # would call run_task() would have to be placed in make_celery() 
    # as well.
    @RPC(validation='named') 
    def launch_task(task_id='', func_name='', args=[], kwargs={}):
        
        match_taskrec = datastore.loadtask(task_id) # Find a matching task record (if any) to the task_id.
        
        if match_taskrec is None: # If we did not find a match...
            if not func_name in task_func_dict: # If the function name is not in the task function dictionary, return an error.
                return_dict = {'error': 'Could not find requested async task function "%s"' % func_name}
            
            else:
                new_task_record = Task(task_id) # Create a new TaskRecord.
                
                # Initialize the TaskRecord with available information.
                new_task_record.status = 'queued'
                new_task_record.queue_time = sc.now()
                new_task_record.func_name = func_name
                new_task_record.args = args
                new_task_record.kwargs = kwargs
                datastore.savetask(new_task_record)  # Add the TaskRecord to the TaskDict.
                
                # Queue up run_task() for Celery.
                my_result = run_task.delay(task_id, func_name, args, kwargs)
                new_task_record.result_id = my_result.id # Add the result ID to the TaskRecord, and update the DataStore.
                datastore.savetask(new_task_record)
                return_dict = new_task_record.jsonify() # Create the return dict from the user repr.
        
        else: # Otherwise (there is a matching task)...
            if match_taskrec.status == 'completed' or match_taskrec.status == 'error': # If the TaskRecord indicates the task has been completed or thrown an error...   
                if match_taskrec.result_id is not None: # If we have a result ID, erase the result from Redis.
                    result = celery_instance.AsyncResult(match_taskrec.result_id)
                    result.forget()
                    match_taskrec.result_id = None
                           
                # Initialize the TaskRecord to start the task again (though possibly with a new function and arguments).
                match_taskrec.status = 'queued'
                match_taskrec.queue_time = sc.now()
                match_taskrec.start_time = None
                match_taskrec.stop_time = None
                match_taskrec.pending_time = None
                match_taskrec.execution_time = None                
                match_taskrec.func_name = func_name
                match_taskrec.args = args
                match_taskrec.kwargs = kwargs
                
                # Queue up run_task() for Celery.   
                my_result = run_task.delay(task_id, func_name, args, kwargs)             
                match_taskrec.result_id = my_result.id # Add the new result ID to the TaskRecord, and update the DataStore.
                datastore.savetask(match_taskrec)
                return_dict = match_taskrec.jsonify() # Create the return dict from the user repr.
            
            else: # Else (the task is not completed)...
                return_dict = {'error': 'Task is already %s' % match_taskrec.status}
        
        return return_dict   # Return our result.
    
    return celery_instance # Return the new instance.


def add_task_funcs(new_task_funcs):
    ''' Function for adding new task functions to those that run_task() can see '''
    global task_func_dict
    for key in new_task_funcs: # For all of the keys in the dict passed in, put the key/value pairs in the global dict.
        task_func_dict[key] = new_task_funcs[key]
    return None


@RPC(validation='named') 
def check_task(task_id, verbose=False): 
    match_taskrec = datastore.loadtask(task_id) # Find a matching task record (if any) to the task_id.
    if match_taskrec is None: # Check to see if the task exists, and if not, return an error.
        errormsg = {'error': 'No task found for specified task ID (%s)' % task_id}
        if verbose: print(errormsg)
        return errormsg
    else: # Update the elapsed times.
        if match_taskrec.pending_time is not None: # If we are no longer pending...
            pending_time = match_taskrec.pending_time # Use the existing pending_time.
            if match_taskrec.execution_time is not None: # If we have finished executing...
                execution_time = match_taskrec.execution_time # Use the execution time in the record.
            else: # Else (we are still executing)...
                execution_time = (sc.now() - match_taskrec.start_time).total_seconds()
        else: # Else (we are still pending)...
            pending_time = (sc.now() - match_taskrec.queue_time).total_seconds()
            execution_time = 0
        taskrec_dict = match_taskrec.jsonify() # Create the return dict from the user repr.
        taskrec_dict['pendingTime'] = pending_time
        taskrec_dict['executionTime'] = execution_time
        if verbose: sc.pp(taskrec_dict)
        return taskrec_dict    # Return the has record information and elapsed times.     
   
   
@RPC(validation='named') 
def get_task_result(task_id):
    match_taskrec = datastore.loadtask(task_id) # Find a matching task record (if any) to the task_id.
    if match_taskrec is None: # Check to see if the task exists, and if not, return an error.
        return {'error': 'No task found for specified task ID'}
    else:
        if match_taskrec.result_id is not None: # If we have a result ID...
            result = celery_instance.AsyncResult(match_taskrec.result_id) # Get the result itself from Celery.
            if not result.ready(): # If the result is not ready, return an error.
                return {'error': 'Task not completed'}
            elif result.failed(): # Else, if we have failed, return the exception.
                return {'error': 'Task failed with an exception'}
            else: # Else (task is ready)...
                return {'result': result.get()}
        else: # Else (no result ID)...
            return {'error': 'No result ID'}
    
    
@RPC(validation='named') 
def delete_task(task_id): 
    match_taskrec = datastore.loadtask(task_id) # Find a matching task record (if any) to the task_id.
    if match_taskrec is None: # Check to see if the task exists, and if not, return an error.
        return {'error': 'No task found for specified task ID'}
    else: # Otherwise (matching task).
        if match_taskrec.result_id is not None: # If we have a result ID, erase the result from Redis.
            result = celery_instance.AsyncResult(match_taskrec.result_id)
            result.revoke(terminate=True) # This commmand works under Linux, but not under the Windows Celery setup.  It allows terminating a task in mid-run.
            result.forget()
        datastore.delete(task_id, objtype='task') # Erase the TaskRecord.
        return 'success'


def taskwrapper(task_func_dict):
    ''' Function for making a register_async_task decorator in other modules '''
    def task_func_decorator(task_func):
        @wraps(task_func)
        def wrapper(*args, **kwargs):        
            output = task_func(*args, **kwargs)
            return output
            
        task_func_dict[task_func.__name__] = task_func # Add the function to the dictionary.
        return wrapper    
    
    return task_func_decorator




__all__ = ['__version__', '__versiondate__', '__license__']

__version__      = '0.17.1'
__versiondate__  = '2022-02-02'
__license__      = 'ScirisWeb %s (%s) -- (c) Sciris.org' % (__version__, __versiondate__)



"""
Server tests. Do not run with pytest since relies on browser support.
"""

import pylab as pl
import sciris as sc
import scirisweb as sw


torun = [
'blank', # An empty figure
'browser', # Simple example of three figures
'advanced', # Illustrates separate legend plotting and direct editing of the JSON
]

if 'doplot' not in locals(): doplot = True


if __name__ == '__main__':

    if 'blank' in torun and doplot:
        if doplot:
            sw.browser()


    if 'browser' in torun:
        figs = []
        for n in [10, 50]:
            fig = pl.figure()
            pl.plot(pl.rand(n), pl.rand(n))
            figs.append(fig)
        barfig = pl.figure()
        pl.bar(pl.arange(10), pl.rand(10))
        barjson = sw.mpld3ify(barfig)
        if doplot:
            sw.browser(figs=figs+[barjson])



    if 'advanced' in torun:

        def make_fig():
            fig = pl.figure()
            ax = fig.add_subplot(111)
            ax.plot([1,4,3,4,], label='mish')
            ax.plot([8,4,3,2], label='mashed potatoes')
            return fig,ax

        fig,ax = make_fig()
        legend = sc.separatelegend(ax, figsettings={'figsize':(4.8,4.8)})

        json1 = sw.mpld3ify(fig,    jsonify=False)
        json2 = sw.mpld3ify(legend, jsonify=False)
        json2['axes'][0]['texts'][1]['text'] = 'mashed potatoes' # Change legend text
        json2['axes'][0]['paths'] = [] # Remove the box around the legend

        if doplot:
            sw.browser([json1, json2])


"""
A simple server used to show mpld3 images -- based on _server in the mpld3 package.

Version: 2019may23
"""

import threading
import webbrowser
import socket
import itertools
import random
import json
#import mpld3 # See below -- temporary bugfix
import pylab as pl
import sciris as sc
try:    import BaseHTTPServer as server # Python 2.x
except: from http import server # Python 3.x

__all__ = ['serve', 'mpld3ify', 'browser'] # Others are for internal use only

def generate_handler(html, files=None):
    if files is None:
        files = {}

    class MyHandler(server.BaseHTTPRequestHandler):
        def do_GET(self):
            """Respond to a GET request."""
            if self.path == '/':
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(html.encode())
            elif self.path in files:
                content_type, content = files[self.path]
                self.send_response(200)
                self.send_header("Content-type", content_type)
                self.end_headers()
                self.wfile.write(content.encode())
            else:
                self.send_error(404)

    return MyHandler


def find_open_port(ip, port, n=50):
    """Find an open port near the specified port"""
    ports = itertools.chain((port + i for i in range(n)), (port + random.randint(-2 * n, 2 * n)))
    for port in ports:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = s.connect_ex((ip, port))
        s.close()
        if result != 0: return port
    raise ValueError("No open ports found")


def serve(html, ip='127.0.0.1', port=8888, n_retries=50):
    """Start a server serving the given HTML, and open a browser

    Example:
       html = '<b>Hello, world!</b>'
       import scirisweb as sw
       sw.serve(html)

    Parameters
    ----------
    html : string
        HTML to serve
    ip : string (default = '127.0.0.1')
        ip address at which the HTML will be served.
    port : int (default = 8888)
        the port at which to serve the HTML
    n_retries : int (default = 50)
        the number of nearby ports to search if the specified port is in use.
    """
    port = find_open_port(ip, port, n_retries)
    Handler = generate_handler(html)

    # Create server
    srvr = server.HTTPServer((ip, port), Handler)

    # Use a thread to open a web browser pointing to the server
    try: browser = webbrowser.get('google-chrome') # CK: Try google Chrome first
    except: browser = webbrowser.get() # If not, just use whatever
    b = lambda: browser.open('http://{0}:{1}'.format(ip, port))
    threading.Thread(target=b).start()

    # CK: don't serve forever, just create it once
    srvr.handle_request()


def mpld3ify(fig, sanitize=True, jsonify=True):
    ''' Do all the processing steps that might be necessary to render a figure displayable '''
    
    import mpld3 # Do this here since currently a bug that sets the wrong backend
    
    # Nothing to do if already a string
    if sc.isstring(fig):
        return fig 
    
    # If it's empty or null, make a small figure
    if not fig: 
        fig = pl.figure(figsize=(1,1)) # Make a dummy plot since no legend
        fig.add_subplot(111) # So axis commands work
        fig.get_axes()[0].set_visible(False) # Turn off axes
    
    # Convert to mpld3 -- this is the big step
    if isinstance(fig, pl.Figure):
        fig = mpld3.fig_to_dict(fig)
    
    # Optionally do additional sanitization
    if sanitize:                         fig = sc.sanitizejson(fig)
    if jsonify and not sc.isstring(fig): fig = json.dumps(fig) # Could cause problems to jsonify a string
    
    return fig


def browser(figs=None, doserve=True, legacy=False, jquery_url=None, d3_url=None, mpld3_url=None):
    ''' 
    Create an MPLD3 GUI and display in the browser.
    
    Usage example:
        import pylab as pl
        figs = []
        for n in [10, 50]:
            fig = pl.figure()
            pl.plot(pl.rand(n), pl.rand(n))
            figs.append(fig)
        qs.browser(figs)
    
    figs can be a single figure or a list of figures.
    
    With doserve=True, launch a web server. Otherwise, return the HTML representation of the figures.
    
    Version: 2018sep08
    '''

    ## Specify the div style, and create the HTML template we'll add the data to -- WARNING, library paths are hardcoded!
    divstyle = "float: left"
    if legacy:
        if jquery_url is None: jquery_url = 'https://code.jquery.com/jquery-1.11.3.min.js'
        if d3_url     is None: d3_url     = 'https://mpld3.github.io/js/d3.v3.min.js'
        if mpld3_url  is None: mpld3_url  = 'https://mpld3.github.io/js/mpld3.v0.3git.js'
    else:
        if jquery_url is None: jquery_url = 'https://code.jquery.com/jquery-1.11.3.min.js'
        if d3_url     is None: d3_url     = 'http://thekerrlab.com/tmp/d3.v5.min.js'
        if mpld3_url  is None: mpld3_url  = 'http://thekerrlab.com/tmp/mpld3.v0.4.1.min.js'
    html = '''
    <html>
    <head><script src="%s"></script></head>
    <body>
    !MAKE DIVS!
    <script>
    function mpld3_load_lib(url, callback){var s = document.createElement('script'); s.src = url; s.async = true; s.onreadystatechange = s.onload = callback; s.onerror = function(){console.warn("failed to load library " + url);}; document.getElementsByTagName("head")[0].appendChild(s)} mpld3_load_lib("%s", function(){mpld3_load_lib("%s", function(){
    !DRAW FIGURES!
    })});
    </script>
    </body></html>
    ''' % (jquery_url, d3_url, mpld3_url)
    
    ## Create the figures to plot
    figjsons = []
    if not figs: figs = sc.emptyfig() # figs is None, or '', or [], etc. -- render a blank figure instead
    figs  = sc.promotetolist(figs)
    for fig in figs: # Loop over each plot
        figjsons.append(mpld3ify(fig)) # Save to JSON
    
    ## Create div and JSON strings to replace the placeholers above
    divstr = ''
    jsonstr = ''
    for f,figjson in enumerate(figjsons):
        divstr += '<div style="%s" id="fig%i" class="fig"></div>\n' % (divstyle, f) # Add div information: key is unique ID for each figure
        jsonstr += 'mpld3.draw_figure("fig%i", %s);\n' % (f, figjson) # Add the JSON representation of each figure -- THIS IS KEY!
    html = html.replace('!MAKE DIVS!',divstr) # Populate div information
    html = html.replace('!DRAW FIGURES!',jsonstr) # Populate figure information
    
    ## Launch a server or return the HTML representation
    if doserve: serve(html)
    else:       return html


"""
rpcs.py -- code related to Sciris RPCs

Notes:
* RPCs are NOT used in cases where there is no (HTML / JavaScript) client 
defined.
* All RPCs are routed to the /api/rpcs endpoint (see rpc-service.js).
    
Last update: 11/11/18 (gchadder3)
"""

# Imports

from functools import wraps
import sciris as sc

__all__ = ['ScirisRPC', 'RPCwrapper']



class ScirisRPC(sc.prettyobj):
    '''
    Validation type:
        'none' : no validation required
        'any':   any login validates
        'named': any non-anonymous user validates
        'admin': any admin login validates
        'user <name>': being logged in as <name> validates (TODO)
        'disabled': presently disabled for clients
    '''
    def __init__(self, call_func, call_type='normal', override=False, validation='none'):
        self.call_func  = call_func
        self.funcname   = call_func.__name__
        self.call_type  = call_type
        self.override   = override
        self.validation = validation
            

        
def RPCwrapper(RPC_dict=None, **callerkwargs):
    def RPC_decorator_factory(**callerkwargs):
        def RPC_decorator(RPC_func):
            @wraps(RPC_func)
            def wrapper(*args, **kwargs):        
                output = RPC_func(*args, **kwargs)
                return output
            RPC_dict[RPC_func.__name__] = ScirisRPC(RPC_func, **callerkwargs) # Create the RPC and add it to the dictionary.
            return wrapper
        return RPC_decorator
    return RPC_decorator_factory


'''
Import the key functions required for the webapp.

Example usage:
    import scirisweb as sw
    app = sw.ScirisApp()

Version: 2020mar13
'''

# Import everything
from .sw_version   import * # analysis:ignore
from .sw_rpcs      import * # analysis:ignore
from .sw_users     import * # analysis:ignore
from .sw_tasks     import * # analysis:ignore
from .sw_datastore import * # analysis:ignore
from .sw_app       import * # analysis:ignore
from .sw_server    import * # analysis:ignore

# Print the license...or not
# print(__license__)


"""
user.py -- code related to Sciris user management
    
Last update: 2018sep20
"""

from flask import Flask, session, current_app as app # analysis:ignore
from flask_login import current_user, login_user, logout_user
import six
import sciris as sc
from . import sw_rpcs as rpcs


##############################################################
### Globals
##############################################################

__all__ = ['RPC_dict']

RPC_dict = {}
RPC = rpcs.RPCwrapper(RPC_dict) # RPC registration decorator factory created using call to make_RPC().



##############################################################
### Classes
##############################################################

__all__ += ['User']


class User(sc.prettyobj):
    '''
    A Sciris user.
                    
    Attributes:
        is_authenticated (bool) -- is this user authenticated? (attribute required by Flask-Login)
        is_active (bool)        -- is this user's account active? (attribute required by Flask-Login)
        is_anonymous (bool)     -- is this user considered anonymous?  (attribute required by Flask-Login)
        is_admin (bool)         -- does this user have admin rights?
        username (str)          -- the username used to log in
        displayname (str)       -- the user's name, which gets displayed in the  browser
        email (str)             -- the user's email     
        password (str)          -- the user's SHA224-hashed password
        raw_password (str)      -- the user's unhashed password
    '''
    
    def  __init__(self, username=None, password=None, displayname=None, email=None, uid=None, raw_password=None, is_admin=False):
        # Handle general properties
        if not username:    username    = 'default'
        if not password:    password    = 'default'
        if not displayname: displayname = username
        if not uid:         uid         = sc.uuid()
        
        # Set user-specific properties
        self.username         = username # Set the username.
        self.displayname      = displayname # Set the displayname (what the browser will show).
        self.email            = email  # Set the user's email.
        self.uid              = uid # The user's UID
        self.is_authenticated = True # Set the user to be authentic.
        self.is_active        = True # Set the account to be active.
        self.is_anonymous     = False # The user is not anonymous.
        self.is_admin         = is_admin # Set whether this user has admin rights.
        
        # Handle the password
        if raw_password is not None:
            if six.PY3: raw_password = raw_password.encode('utf-8')
            password = sc.sha(raw_password).hexdigest()
        self.password = password
        return None
    
    
    def get_id(self):
        ''' Method required by Flask-login '''
        return self.username
    
    
    def show(self, verbose=False):
        ''' Display the user's properties '''
        if not verbose: # Simply display the username and UID
            print('Username: %s; UID: %s' % (self.username, self.uid))
        else: # Or full information
            print('---------------------')
            print('        Username: %s' % self.username)
            print('    Display name: %s' % self.displayname)
            print(' Hashed password: %s' % self.password)
            print('   Email address: %s' % self.email)
            print('             UID: %s' % self.uid)
            print('Is authenticated: %s' % self.is_authenticated)
            print('       Is active: %s' % self.is_active)
            print('    Is anonymous: %s' % self.is_anonymous)
            print('        Is admin: %s' % self.is_admin)
            print('---------------------')
        return None
    
    
    def jsonify(self, verbose=False):
        ''' Return a JSON-friendly representation of a user '''
        output = {'user':
                         {'username':    self.username, 
                          'displayname': self.displayname, 
                          'email':       self.email,
                          'uid':         self.uid}
                 }
        if verbose:
            output['user'].update({'is_authenticated': self.is_authenticated,
                                   'is_active':        self.is_active,
                                   'is_anonymous':     self.is_anonymous,
                                   'is_admin':         self.is_admin,
                                   'created':          self.created,
                                   'modified':         self.modified[-1]})
        return output



##############################################################
### Functions and RPCs
##############################################################

__all__ += ['save_user', 'load_user', 'user_login', 'user_logout', 'user_register']
__all__ += ['user_change_info', 'user_change_password', 'admin_delete_user', 'admin_activate_account']
__all__ += ['admin_deactivate_account', 'admin_grant_admin', 'admin_revoke_admin', 'admin_reset_password', 'make_default_users']



def save_user(user):
    ''' Save the specified user to the DataStore '''
    app.datastore.saveuser(user)
    return None


def load_user(username=None):
    ''' Load the currently active user from the DataStore '''
    if username is None:
        try: 
            username = current_user.get_id()
        except Exception as E:
            errormsg = 'No username supplied and could not get current user: %s' % str(E)
            raise Exception(errormsg)
    user = app.datastore.loaduser(username)
    return user


@RPC()
def user_login(username, password, verbose=False):  
    matching_user = app.datastore.loaduser(username, die=False) # Get the matching user (if any).
    
    # If we have a match and the password matches, and the account is active, also, log in the user and return success; otherwise, return failure.
    if matching_user and matching_user.password==password and matching_user.is_active:
        if verbose: print('User %s found, logging in' % username)
        login_user(matching_user) # Log the user in.
        matching_user.show(verbose=verbose)
        return 'success'
    else:
        if   not matching_user:                    errormsg = 'Login failed: user "%s" does not exist.' % username
        elif not matching_user.password==password: errormsg = 'Login failed: password for user "%s" is incorrect.' % username
        elif not matching_user.is_active:          errormsg = 'Login failed: user "%s" is inactive.' % username
        else:                                      errormsg = 'Unknown error'
        if verbose: print('User %s not found, not logging in (%s)' % (username, errormsg))
        return errormsg
    
    
@RPC(validation='named')
def user_logout():
    logout_user() # Log the user out and set the session to having an anonymous user.
    session.clear() # Clear the session cookie.
    return None # Return nothing.


@RPC()
def user_register(username, password, displayname, email): 
    matching_user = app.datastore.loaduser(username, die=False) # Get the matching user (if any).
    if matching_user is not None: # If we have a match, fail because we don't want to register an existing user.
        errormsg = 'User registration failed: user "%s" already exists' % username
        return errormsg
    new_user = User(username=username, password=password, displayname=displayname, email=email) # Create a new User object with the new information.
    new_user.is_active = app.config['REGISTER_AUTOACTIVATE'] # Optionally set the user to be inactive (so an admin user needs to activate it)
    app.datastore.saveuser(new_user) # Save the user
    return 'success' # Return success.


@RPC(validation='named')
def user_change_info(username, password, displayname, email):
    the_user = app.datastore.loaduser(current_user.username) # Reload the current user from the database
    if password != the_user.password: # If the password entered doesn't match the current user password, fail.
        errormsg = 'User info change failed: password for user "%s" is incorrect.' % username
        return errormsg
       
    # If the username entered by the user is different from the current_user name (meaning they are trying to change the name)...
    if username != the_user.username:
        matching_user = app.datastore.loaduser(username) # Get any matching user (if any) to the new username we're trying to switch to.
        if matching_user is not None: # If we have a match, fail because we don't want to rename the user to another existing user.
            errormsg = 'User rename failed: user "%s" already exists.' % username
            return errormsg
        
    # Change the user name, display name, email, and instance label.
    the_user.username    = username
    the_user.displayname = displayname
    the_user.email       = email
    
    # Update the user in user_dict.
    app.datastore.saveuser(the_user, overwrite=True)
    return 'success'


@RPC(validation='named') 
def user_change_password(oldpassword, newpassword):
    the_user = app.datastore.loaduser(current_user.username) # Reload the current user from the database
    
    # If the password entered doesn't match the current user password, fail.
    if oldpassword != the_user.password:
        errormsg = 'Password change failed: password for user "%s" is incorrect.' % the_user.username
        return errormsg
    
    the_user.password = newpassword # Change just the password.
    app.datastore.saveuser(the_user, overwrite=True) # Update the user in user_dict.
    return 'success'


@RPC(validation='admin')
def admin_delete_user(username):
    matching_user = app.datastore.loaduser(username) # Get the matching user (if any).
    if matching_user is None: return 'failure' # If we don't have a match, fail.
    app.datastore.delete(objtype='user', uid=username) # Delete the user from the dictionary.
    return 'success'


@RPC(validation='admin')
def admin_activate_account(username):
    # Get the matching user (if any).
    matching_user = app.datastore.loaduser(username)
    
    # If we don't have a match, fail.
    if matching_user is None:
        return 'failure'
    
    # If the account is already activated, fail.
    if matching_user.is_active:
        return 'failure'
    
    # Activate the account.
    matching_user.is_active = True
    app.datastore.saveuser(matching_user, overwrite=True)
    
    # Return success.
    return 'success'    

@RPC(validation='admin')
def admin_deactivate_account(username):
    # Get the matching user (if any).
    matching_user = app.datastore.loaduser(username)
    
    # If we don't have a match, fail.
    if matching_user is None:
        return 'failure'
    
    # If the account is already deactivated, fail.
    if not matching_user.is_active:
        return 'failure'
    
    # Activate the account.
    matching_user.is_active = False
    app.datastore.saveuser(matching_user, overwrite=True)
    
    # Return success.
    return 'success'

@RPC(validation='admin')
def admin_grant_admin(username):
    # Get the matching user (if any).
    matching_user = app.datastore.loaduser(username)
    
    # If we don't have a match, fail.
    if matching_user is None:
        return 'failure'
    
    # If the account already has admin access, fail.
    if matching_user.is_admin:
        return 'failure'
    
    # Grant admin access.
    matching_user.is_admin = True
    app.datastore.saveuser(matching_user, overwrite=True)
    
    # Return success.
    return 'success'

@RPC(validation='admin')
def admin_revoke_admin(username):
    # Get the matching user (if any).
    matching_user = app.datastore.loaduser(username)
    
    # If we don't have a match, fail.
    if matching_user is None:
        return 'failure'
    
    # If the account has no admin access, fail.
    if not matching_user.is_admin:
        return 'failure'
    
    # Revoke admin access.
    matching_user.is_admin = False
    app.datastore.saveuser(matching_user, overwrite=True)
    
    # Return success.
    return 'success'

@RPC(validation='admin')
def admin_reset_password(username):
    # Get the matching user (if any).
    matching_user = app.datastore.loaduser(username)
    
    # If we don't have a match, fail.
    if matching_user is None:
        return 'failure'
    
    # Set the password to the desired raw password for them to use.
    raw_password = 'sciris'
    matching_user.password = sc.sha(raw_password).hexdigest() 
    app.datastore.saveuser(matching_user, overwrite=True)
    
    # Return success.
    return 'success'


@RPC(validation='named') 
def get_current_user_info():
    return current_user.jsonify()


def make_default_users(app, include_admin=False):
    # Create two test Users that can get added
    test_user = User(username='demo', raw_password='demo', uid='abcdef0123456789')
    users = [test_user]
    if include_admin:
        admin_user = User(username='admin', raw_password='admin', uid='00112233445566778899', is_admin=True)
        users.append(admin_user)
    for user in users:
        app.datastore.saveuser(user, overwrite=False, die=False) # Save the user
    return users


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 16:57:44 2019

@author: cliffk
"""

from distributed import Scheduler
from tornado.ioloop import IOLoop
from threading import Thread

loop = IOLoop.current()
t = Thread(target=loop.start, daemon=True)
t.start()

s = Scheduler(loop=loop)
s.start('tcp://:8786')   # Listen on TCP port 8786



'''
Super simple Sciris app.

To run, simply type

   python app.py

after ScirisWeb has been installed.

To use, go to localhost:8080 in your browser.

Version: 2019jan12
'''

import pylab as pl
import scirisweb as sw
app = sw.ScirisApp(__name__, name="helloworld")

@app.route("/")
def hello():
    output = 'Hello world!<br><br>Click <a href="/number">here</a> for a random number.'
    return output

@app.route("/number")
def number():
    number = pl.rand()
    output = 'Random number: %s.<br><br>Click <a href="/">here</a> to go back.' % number
    return output

# Run the webapp
if __name__ == "__main__":
    app.run(with_client=False)


"""
sw_app.py -- classes for Sciris (Flask-based) apps 
    
Last update: 2018nov14
"""

# Imports
import io
import os
import sys
import socket
import logging
import traceback
from collections import OrderedDict
from functools import wraps

from flask import Flask, request, abort, json, jsonify as flask_jsonify, send_from_directory, make_response, current_app as flaskapp, send_file
from flask_login import LoginManager, current_user
from twisted.internet import reactor
from twisted.internet.endpoints import serverFromString
from twisted.logger import globalLogBeginner, FileLogObserver, formatEvent
from twisted.python.threadpool import ThreadPool
from twisted.web.resource import Resource
from twisted.web.server import Site
from twisted.web.static import File
from twisted.web.wsgi import WSGIResource
from werkzeug.serving import run_with_reloader
from werkzeug.exceptions import HTTPException
from werkzeug.utils import secure_filename

import sciris as sc
from . import sw_datastore as ds
from . import sw_rpcs as rpcs
from . import sw_tasks as tasks
from . import sw_users as users

#################################################################
### Classes and functions
#################################################################

__all__ = ['robustjsonify', 'ScirisApp', 'ScirisResource', 'flaskapp']

def robustjsonify(response, fallback=False, verbose=True):
    ''' 
    Flask's default jsonifier clobbers dict order; this preserves it; however, it falls 
    back to regular Flask jsonify if needed.
    '''
    robustjson = sc.sanitizejson(response, tostring=True) + '\n' # The newline is part of Flask: https://github.com/pallets/flask/issues/1877
    lenjson = len(robustjson)
    placeholder = '_'*(lenjson-3) # Create a placeholder of the correct length -- -3 because of ""\n which gets added
    flaskjson = flask_jsonify(placeholder) # Use standard Flask jsonification
    lenplaceholder = len(flaskjson.response[0])
    if lenplaceholder == lenjson: # Lengths match, proceed
        flaskjson.response[0] = robustjson 
    else: # If there's a length mismatch, fall back to regular Flask
        if verbose: print('ScirisApp: not robustifying JSON since length mismatch (%s vs. %s): %s' % (lenjson, len(flaskjson.response[0]), robustjson))
        fallback = True
    if fallback: # Use standard Flask jsonification if anything went wrong
        try:
            flaskjson = flask_jsonify(sc.sanitizejson(response)) 
        except Exception as E:
            errormsg = 'Flask jsonification of "%s" failed: %s' % (response, str(E))
            raise Exception(errormsg)
    return flaskjson


class ScirisApp(sc.prettyobj):
    """
    An object encapsulating a Sciris webapp, generally.  This app has an 
    associated Flask app that actually runs to listen for and respond to 
    HTTP requests.
    
    Methods:
        __init__(script_path: str, self.config: config module [None], 
            client_dir: str [None]): void -- constructor
        run(with_twisted: bool [True], with_flask: bool [True], 
            with_client: [True]): void -- run the actual server
        define_endpoint_layout(rule: str, layout: list): void -- set up an 
            endpoint with a layout of a static Flask page
        add_RPC(new_RPC: ScirisRPC): void -- add an RPC to the app's dictionary
        add_RPC_dict(new_RPC_dict: dict): void -- add RPCs from another RPC 
            dictionary to the app's dictionary
        register_RPC(**kwargs): func -- decorator factory for adding a function 
            as an RPC
        _layout_render(): void -- render HTML for the layout of the given 
            rule (in the request)
        _do_RPC(): void -- process a request in such a way as to find and 
            dispatch the chosen RPC, doing any validation checking and error 
            logging necessary in the process
                    
    Attributes:
        flask_app (Flask) -- the actual Flask app
        config (flask.config.Config) -- the Flask configuration dictionary
        define_endpoint_callback (func) -- points to the flask_app.route() 
            function so you can use @app.define_endpoint_callback in the calling code
        endpoint_layout_dict (dict) -- dictionary holding static page layouts
        RPC_dict (dict) -- dictionary of site RPCs
            
    Usage:
        >>> app = ScirisApp(__file__)                      
    """
    
    def  __init__(self, filepath=None, config=None, name=None, RPC_dict=None, logfile=None, colorize=None, **kwargs):

        # If we're sending output to a logfile, initialize it at the very beginning
        if logfile:
            self.stdout = sys.stdout # Save the original so we can restore it if need be
            try:
                print('Redirecting output to %s' % logfile)
                sys.stdout = open(logfile, 'a+', 1) # Open file for appending, with buffer size of 1
            except Exception as E:
                errormsg = 'Could not open logfile "%s": %s' % (logfile, str(E))
                raise Exception(errormsg)
        
        # Decide whether to use colorization -- yes unless a logfile is being used
        if colorize is None:
            colorize = False if logfile else True
        self.colorize= colorize

        # Initialize everything else
        if name is None: 
            name = 'default'
        self.name = name
        self.flask_app = Flask(__name__) # Open a new Flask app.
        if config is not None: # If we have a config module, load it into Flask's config dict.
            self.flask_app.config.from_object(config)
        self.config = self.flask_app.config # Set an easier link to the configs dictionary.
        self.config['ROOT_ABS_DIR'] = os.path.dirname(os.path.abspath(filepath)) # Extract the absolute directory path from the above.
        self._init_logger() # Initialize the Flask logger. 
        self._set_config_defaults() # Set up default values for configs that are not already defined.
        self._update_config_defaults(**kwargs) # If additional command-line arguments are supplied, use them
        self.define_endpoint_callback = self.flask_app.route # Set an alias for the decorator factory for adding an endpoint.
        self.endpoint_layout_dict = {} # Create an empty layout dictionary.
        self.RPC_dict = {}  # Create an empty RPC dictionary.
        
        # Set up file paths.
        self._init_file_dirs()
        
        # Set up RPCs
        if RPC_dict:
            self.add_RPC_dict(RPC_dict)
        
        # If we are including DataStore functionality, initialize it.
        if self.config['USE_DATASTORE']:
            self._init_datastore(use_db=True)
            self.flask_app.datastore = self.datastore
        else: # Initialize to be a temporary folder
            self._init_datastore(use_db=False)

        # If we are including DataStore and users functionality, initialize users.
        if self.config['USE_DATASTORE'] and self.config['USE_USERS']:
            self.login_manager = LoginManager() # Create a LoginManager() object.
            
            # This function gets called when authentication gets done by 
            # Flask-Login.  userid is the user ID pulled out of the session 
            # cookie the browser passes in during an HTTP request.  The 
            # function returns the User object that matches this, so that the 
            # user data may be used to authenticate (for example their 
            # rights to have admin access).  The return sets the Flask-Login 
            # current_user value.
            @self.login_manager.user_loader
            def load_user(userid):
                return self.datastore.loaduser(userid) # Return the matching user (if any).
            
            self.login_manager.init_app(self.flask_app) # Configure Flask app for login with the LoginManager.
            self.add_RPC_dict(users.RPC_dict) # Register the RPCs in the users.py module.
            
        # If we are including DataStore and tasks, initialize them.    
        if self.config['USE_DATASTORE'] and self.config['USE_TASKS']:
            self._init_tasks() # Initialize the users.
            self.add_RPC_dict(tasks.RPC_dict) # Register the RPCs in the user.py module.

        return None # End of __init__
            
    def _init_logger(self):
        self.flask_app.logger.setLevel(logging.DEBUG)
        return None
    
    def _set_config_defaults(self):
        if 'CLIENT_DIR'         not in self.config: self.config['CLIENT_DIR']         = '.'        
#        if 'CLIENT_DIR'         not in self.config: self.config['CLIENT_DIR']         = None
        if 'LOGGING_MODE'       not in self.config: self.config['LOGGING_MODE']       = 'FULL' 
        if 'SERVER_PORT'        not in self.config: self.config['SERVER_PORT']        = 8080
        if 'USE_DATASTORE'      not in self.config: self.config['USE_DATASTORE']      = False
        if 'USE_USERS'          not in self.config: self.config['USE_USERS']          = False
        if 'USE_TASKS'          not in self.config: self.config['USE_TASKS']          = False
        if 'MATPLOTLIB_BACKEND' not in self.config: self.config['MATPLOTLIB_BACKEND'] = 'Agg'
        if 'SLACK'              not in self.config: self.config['SLACK']              = None
        return None
    
    def _update_config_defaults(self, **kwargs):
        ''' Used to update config with command-line arguments '''
        for key,val in kwargs.items():
            KEY = key.upper() # Since they're always uppercase
            if KEY in self.config:
                origval = self.config[KEY]
                self.config[KEY] = val
                print('Resetting configuration option "%s" from "%s" to "%s"' % (KEY, origval, val))
            else:
                warningmsg = '\nWARNING: kwarg "%s":"%.100s" will be ignored since it is not in the list of valid config options:\n' % (KEY, val)
                for validkey in sorted(self.config.keys()):
                    warningmsg += '  %s\n' % validkey
                print(warningmsg)
        return None
                

    def _init_file_dirs(self):
        # Set the absolute client directory path.
        
        # If we have a client directory...
        if self.config['CLIENT_DIR'] is not None:
            # If we do not have an absolute directory, tack what we have onto the 
            # ROOT_ABS_DIR setting.
            if not os.path.isabs(self.config['CLIENT_DIR']):
                self.config['CLIENT_DIR'] = os.path.join(self.config['ROOT_ABS_DIR'], self.config['CLIENT_DIR'])
        
        return None
        
    def _init_datastore(self, use_db=True):
        if use_db:
            # Create the DataStore object
            self.datastore = ds.make_datastore(url=self.config['DATASTORE_URL'])
            
            if self.config['LOGGING_MODE'] == 'FULL':
                maxkeystoshow = 20
                keys = self.datastore.keys()
                nkeys = len(keys)
                keyinds = range(1,nkeys+1)
                keypairs = list(zip(keyinds, keys))
                print('>> Loaded DataStore with %s key(s)' % nkeys)
                if nkeys>2*maxkeystoshow:
                    print('>> First and last %s keys:' % maxkeystoshow)
                    keypairs    = keypairs[:maxkeystoshow] + keypairs[-maxkeystoshow:]
                for k,key in keypairs:
                    print('  Key %02i: %s' % (k,key))
        else:
            self.datastore = ds.DataDir() # Initialize with a simple temp data directory instead
        return None
    
    def _init_tasks(self):
        # Have the tasks.py module make the Celery app to connect to the worker, passing in the config parameters.
        print('Making Celery instance...')
        tasks.make_celery(self.config)
        
    def run(self, with_twisted=True, with_flask=True, with_client=True, do_log=False, show_logo=True, autoreload=False, run_args=None):
        
        def run_twisted(port=8080, flask_app=None, client_dir=None, do_log=False, reactor_args=None):
            # Give an error if we pass in no Flask server or client path.
            if reactor_args is None:
                reactor_args = {}
        
            if (flask_app is None) and (client_dir is None): 
                print('ERROR: Neither client or server are defined.')
                return None
            if do_log: # Set up logging.
                globalLogBeginner.beginLoggingTo([FileLogObserver(sys.stdout, lambda _: formatEvent(_) + "\n")])
            if client_dir is not None: # If there is a client path, set up the base resource.
                base_resource = File(client_dir)
                
            # If we have a flask app...
            if flask_app is not None:
                thread_pool = ThreadPool(maxthreads=30) # Create a thread pool to use with the app.
                wsgi_app = WSGIResource(reactor, thread_pool, flask_app) # Create the WSGIResource object for the flask server.
                if client_dir is None: # If we have no client path, set the WSGI app to be the base resource.
                    base_resource = ScirisResource(wsgi_app)
                else:  # Otherwise, make the Flask app a child resource.
                    base_resource.putChild(b'api', ScirisResource(wsgi_app))
                thread_pool.start() # Start the threadpool now, shut it down when we're closing
                reactor.addSystemEventTrigger('before', 'shutdown', thread_pool.stop)
        
            # Create the site.
            site = Site(base_resource) 
            endpoint = serverFromString(reactor, "tcp:port=" + str(port)) # Create the endpoint we want to listen on, and point it to the site.
            endpoint.listen(site)
            reactor.run(**reactor_args) # Start the reactor.
            return None
        
        # To allow arguments to be passed to the run function
        if run_args is None:
            run_args = {}
        
        # Initialize plotting
        try:
            import matplotlib.pyplot as ppl # Place here so as not run on import
            ppl.switch_backend(self.config['MATPLOTLIB_BACKEND'])
            print('Matplotlib backend switched to "%s"' % (self.config['MATPLOTLIB_BACKEND']))
        except Exception as E:
            print('Switching Matplotlib backend to "%s" failed: %s' % (self.config['MATPLOTLIB_BACKEND'], repr(E)))
        
        # Display the logo
        port = int(self.config['SERVER_PORT']) # Not sure if casting to int is necessary
        appstring = 'ScirisApp "%s" is now running on port %s' % (self.name, port)
        borderstr = '='*len(appstring)
        logostr = '''\
      ___  ___    %s 
     / __|/ __|   %s     
     \__ \ |__    %s     
     |___/\___|   %s     
                  %s''' % (' '*(len(appstring)+4), borderstr, appstring, borderstr, ' '*(len(appstring)+5))
        logocolors = ['gray','bgblue'] # ['gray','bgblue']
        if show_logo:
            print('')
            for linestr in logostr.splitlines():
                sc.colorize(logocolors, linestr, enable=self.colorize)
            print('')


        if not with_twisted: # If we are not running the app with Twisted, just run the Flask app.
            run_fcn = lambda: self.flask_app.run(port=port, **run_args)
        else:
            twisted_args = {}
            twisted_args['reactor_args'] = run_args
            twisted_args['do_log'] = do_log
            twisted_args['port'] = port
            if with_client:
                twisted_args['client_dir'] = self.config['CLIENT_DIR']
            if with_flask:
                twisted_args['flask_app'] = self.flask_app
            if autoreload:
                twisted_args['reactor_args']['installSignalHandlers'] = 0
            run_fcn = lambda: run_twisted(**twisted_args)

        if autoreload:
            run_fcn = run_with_reloader(run_fcn)

        return run_fcn()
    
    def define_endpoint_layout(self, rule, layout):
        # Save the layout in the endpoint layout dictionary.
        self.endpoint_layout_dict[rule] = layout
        
        # Set up the callback, to point to the _layout_render() function.
        self.flask_app.add_url_rule(rule, 'layout_render', self._layout_render)
        return None
        
    def slacknotification(self, message=None):
        ''' Send a message on Slack '''
        if self.config.get('SLACK'):
            slack_webhook = self.config['SLACK'].get('webhook')
            slack_to      = self.config['SLACK'].get('to')
            slack_from    = self.config['SLACK'].get('from')
            sc.slacknotification(message=message, webhook=slack_webhook, to=slack_to, fromuser=slack_from, die=False)
        else:
            print('Cannot send Slack message "%s": Slack not enabled in config file' % message)
        return None
    
    def route(self, rule, methods=None, *args, **kwargs):
        ''' Shortcut to Flask route decorator '''
        if methods is None: methods = ['GET', 'POST']
        return self.flask_app.route(rule, methods=methods, *args, **kwargs)

    def add_RPC(self, new_RPC):
        # If we are setting up our first RPC, add the actual endpoint.
        if len(self.RPC_dict) == 0:
            self.flask_app.add_url_rule('/rpcs', 'do_RPC', self._do_RPC, methods=['POST'])
          
        # If the function name is in the dictionary...
        if new_RPC.funcname in self.RPC_dict:
            # If we have the power to override the function, give a warning.
            if new_RPC.override:
                if self.config['LOGGING_MODE'] == 'FULL':
                    print('>> add_RPC(): WARNING: Overriding previous version of %s:' % new_RPC.funcname)
                    print('>>   Old: %s.%s' % 
                        (self.RPC_dict[new_RPC.funcname].call_func.__module__, 
                        self.RPC_dict[new_RPC.funcname].call_func.__name__))
                    print('>>   New: %s.%s' % (new_RPC.call_func.__module__, 
                        new_RPC.call_func.__name__))
            # Else, give an error, and exit before the RPC is added.
            else:
                print('>> add_RPC(): ERROR: Attempt to override previous version of %s: %s.%s' % \
                      (new_RPC.funcname, self.RPC_dict[new_RPC.funcname].call_func.__module__, self.RPC_dict[new_RPC.funcname].funcname))
                return
        
        # Create the RPC and add it to the dictionary.
        self.RPC_dict[new_RPC.funcname] = new_RPC
    
    def add_RPC_dict(self, new_RPC_dict):
        for RPC_funcname in new_RPC_dict:
            self.add_RPC(new_RPC_dict[RPC_funcname])

    def register_RPC(self, **callerkwargs):
        def RPC_decorator(RPC_func):
            @wraps(RPC_func)
            def wrapper(*args, **kwargs):
                output = RPC_func(*args, **kwargs)
                return output

            # Create the RPC and try to add it to the dictionary.
            new_RPC = rpcs.ScirisRPC(RPC_func, **callerkwargs)
            self.add_RPC(new_RPC)
            
            return wrapper

        return RPC_decorator

    def _layout_render(self):
        render_str = '<html>'
        render_str += '<body>'
        for layout_comp in self.endpoint_layout_dict[str(request.url_rule)]:
            render_str += layout_comp.render()
        render_str += '</body>'
        render_str += '</html>'
        return render_str
    
    def _do_RPC(self, verbose=False):
        # Check to see whether the RPC is getting passed in in request.form.
        # If so, we are doing an upload, and we want to download the RPC 
        # request info from the form, not request.data.
        if 'funcname' in request.form: # Pull out the function name, args, and kwargs
            fn_name = request.form.get('funcname')
            try:    args = json.loads(request.form.get('args', "[]"), object_pairs_hook=OrderedDict)
            except: args = [] # May or may not be present
            try:    kwargs = json.loads(request.form.get('kwargs', "{}"), object_pairs_hook=OrderedDict)
            except: kwargs = {} # May or may not be present
        else: # Otherwise, we have a normal or download RPC, which means we pull the RPC request info from request.data.
            reqdict = json.loads(request.data, object_pairs_hook=OrderedDict)
            fn_name = reqdict['funcname']
            args = reqdict.get('args', [])
            kwargs = reqdict.get('kwargs', {})
        if verbose:
            print('RPC(): RPC properties:')
            print('  fn_name: %s' % fn_name)
            print('     args: %s' % args)
            print('   kwargs: %s' % kwargs)
        
        # If the function name is not in the RPC dictionary, return an error.
        if not sc.isstring(fn_name) :
            return robustjsonify({'error': 'Invalid RPC - must be a string (%s)' % fn_name})

        if not fn_name in self.RPC_dict:
            return robustjsonify({'error': 'Could not find requested RPC "%s"' % fn_name})
        
        found_RPC = self.RPC_dict[fn_name] # Get the RPC we've found.
        
        ## Do any validation checks we need to do and return errors if they don't pass.
        
        # If the RPC is disabled, always return a Status 403 (Forbidden)
        if found_RPC.validation == 'disabled':
            if verbose: print('RPC(): RPC disabled')
            abort(403)
        
        # Only do other validation if DataStore and users are included -- NOTE: Any "unknown" validation values are treated like 'none'.
        if self.config['USE_DATASTORE'] and self.config['USE_USERS']:
            if found_RPC.validation == 'any' and not (current_user.is_anonymous or current_user.is_authenticated):
                abort(401) # If the RPC should be executable by any user, including an anonymous one, but there is no authorization or anonymous login, return a Status 401 (Unauthorized)
            elif found_RPC.validation == 'named' and (current_user.is_anonymous or not current_user.is_authenticated):
                abort(401) # Else if the RPC should be executable by any non-anonymous user, but there is no authorization or there is an anonymous login, return a Status 401 (Unauthorized)
            elif found_RPC.validation == 'admin': # Else if the RPC should be executable by any admin user, but there is no admin login or it's an anonymous login...
                if current_user.is_anonymous or not current_user.is_authenticated:
                    abort(401) # If the user is anonymous or no authenticated user is logged in, return Status 401 (Unauthorized).
                elif not current_user.is_admin:
                    abort(403) # Else, if the user is not an admin user, return Status 403 (Forbidden).
                    
        # If we are doing an upload...
        if found_RPC.call_type == 'upload':
            if verbose: print('Starting upload...')
            thisfile = request.files['uploadfile'] # Grab the formData file that was uploaded.    
            filename = secure_filename(thisfile.filename) # Extract a sanitized filename from the one we start with.
            try:
                uploaded_fname = os.path.join(self.datastore.tempfolder, filename) # Generate a full upload path/file name.
            except Exception as E:
                errormsg = 'Could not create filename for uploaded file: %s' % str(E)
                raise Exception(errormsg)
            try:
                thisfile.save(uploaded_fname) # Save the file to the uploads directory
            except Exception as E:
                errormsg = 'Could not save uploaded file: %s' % str(E)
                raise Exception(errormsg)
            args.insert(0, uploaded_fname) # Prepend the file name to the args list.
        
        # Show the call of the function.
        callcolor    = ['cyan',  'bgblue']
        successcolor = ['green', 'bgblue']
        failcolor    = ['gray',  'bgred']
        timestr = '[%s]' % sc.now(astype='str')
        try:    userstr = ' <%s>' % current_user.username
        except: userstr =' <no user>'
        RPCinfo = sc.objdict({'time':timestr, 'user':userstr, 'module':found_RPC.call_func.__module__, 'name':found_RPC.funcname})
        
        if self.config['LOGGING_MODE'] == 'FULL':
            string = '%s%s RPC called: "%s.%s"' % (RPCinfo.time, RPCinfo.user, RPCinfo.module, RPCinfo.name)
            sc.colorize(callcolor, string, enable=self.colorize)
    
        # Execute the function to get the results, putting it in a try block in case there are errors in what's being called. 
        try:
            if verbose: print('RPC(): Starting RPC...')
            T = sc.tic()
            result = found_RPC.call_func(*args, **kwargs)
            if isinstance(result, dict) and 'error' in result: # If the RPC returns an error, return it
                return robustjsonify({'error':result['error']})
            elapsed = sc.toc(T, output=True)
            if self.config['LOGGING_MODE'] == 'FULL':
                string = '%s%s RPC finished in %0.2f s: "%s.%s"' % (RPCinfo.time, RPCinfo.user, elapsed, RPCinfo.module, RPCinfo.name)
                sc.colorize(successcolor, string, enable=self.colorize)
        except Exception as E:
            if verbose: print('RPC(): Exception encountered...')
            shortmsg = str(E)
            exception = traceback.format_exc() # Grab the trackback stack
            hostname = '|%s| ' % socket.gethostname()
            tracemsg = '%s%s%s Exception during RPC "%s.%s" \nRequest: %s \n%.10000s' % (hostname, RPCinfo.time, RPCinfo.user, RPCinfo.module, RPCinfo.name, request, exception)
            sc.colorize(failcolor, tracemsg, enable=self.colorize) # Post an error to the Flask logger limiting the exception information to 10000 characters maximum (to prevent monstrous sqlalchemy outputs)
            if self.config['SLACK']:
                self.slacknotification(tracemsg)
            if isinstance(E, HTTPException): # If we have a werkzeug exception, pass it on up to werkzeug to resolve and reply to.
                raise E
            code = 500 # Send back a response with status 500 that includes the exception traceback.
            fullmsg = shortmsg + '\n\nException details:\n' + tracemsg
            reply = {'exception':fullmsg} # NB, not sure how to actually access 'traceback' on the FE, but keeping it here for future
            return make_response(robustjsonify(reply), code)

        # If we are doing a download, prepare the response and send it off.
        if found_RPC.call_type == 'download':
            # To download a file, use `this.$sciris.download` instead of `this.$sciris.rpc`. Decorate the RPC with
            # `@RPC(call_type='download')`. Finally, the RPC needs to specify the file and optionally the filename.
            # This is done with tuple unpacking. The following outputs are supported from `rpc_function()`
            #
            # 1 - filename_on_disk
            # 2 - BytesIO
            # 3 - filename_on_disk, download_filename
            # 4- BytesIO, download_filename
            #
            # Examples return values from the RPC are as follows
            #
            # 1 - "E:/test.xlsx" (uses "test.xlsx")
            # 2 - <BytesIO> (default filename will be generated in this function)
            # 3 - ("E:/test.xlsx","foo.xlsx")
            # 4 - (<BytesIO>,"foo.xlsx")
            #
            # On the RPC end, the most common cases would be it might look like
            #
            # return "E:/test.xlsx"
            #
            # OR
            #
            # return Blobject.to_file(), "foo.xlsx"

            if verbose: print('RPC(): Starting download...')

            if result is None: # If we got None for a result (the full file name), return an error to the client.
                return robustjsonify({'error': 'Could not find resource to download from RPC "%s": result is None' % fn_name})
            elif sc.isstring(result):
                from_file = True
                dir_name, file_name = os.path.split(result)
                output_name = file_name
            elif isinstance(result,io.BytesIO):
                from_file = False
                bytesio = result
                output_name = 'download.obj'
            else:
                try:
                    content = result[0]
                    output_name = result[1]
                    if sc.isstring(content):
                        from_file = True
                        dir_name, file_name = os.path.split(content)
                    elif isinstance(content,io.BytesIO):
                        from_file = False
                        bytesio = content
                    else:
                        return robustjsonify({'error': 'Unrecognized RPC output'})
                except Exception as E:
                    return robustjsonify({'error': 'Error reading RPC result (%s)' % E})

            if from_file:
                response = send_from_directory(dir_name, file_name, as_attachment=True)
                response.status_code = 201  # Status 201 = Created
                # Unfortunately, we cannot remove the actual file at this point
                # because it is in use during the actual download, so we rely on
                # later cleanup to remove download files.
            else:
                response = send_file(bytesio, as_attachment=True, attachment_filename=output_name)
            response.headers['filename'] = output_name
            print(response)
            return response # Return the response message.


        # Otherwise (normal and upload RPCs), 
        else: 
            if found_RPC.call_type == 'upload': # If we are doing an upload....
                try:
                    os.remove(uploaded_fname) # Erase the physical uploaded file, since it is no longer needed.
                    if verbose: print('RPC(): Removed uploaded file: %s' % uploaded_fname)
                except Exception as E:
                    if verbose: print('RPC(): Could not remove uploaded file: %s' % str(E)) # Probably since moved by the user
            if result is None: # If None was returned by the RPC function, return ''.
                if verbose: print('RPC(): RPC finished, returning None')
                return ''
            else: # Otherwise, convert the result (probably a dict) to JSON and return it.
                output = robustjsonify(result)
                if verbose: print('RPC(): RPC finished, returning result')
                return output
        
        
class ScirisResource(Resource):
    isLeaf = True

    def __init__(self, wsgi):
        self._wsgi = wsgi

    def render(self, request):
        r = self._wsgi.render(request) # Get the WSGI render results (i.e. for Flask app).

        # Keep the client browser from caching Flask response, and set the response as already being "expired."
        request.responseHeaders.setRawHeaders(b'Cache-Control', [b'no-cache', b'no-store', b'must-revalidate'])
        request.responseHeaders.setRawHeaders(b'expires', [b'0'])
        
        # Pass back the WSGI render results.
        return r
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 17:18:08 2019

@author: cliffk
"""

from dask import compute, delayed
import pylab as pl
import sciris as sc

from distributed import LocalCluster
c = LocalCluster(processes=False)

print(c.scheduler)
print(c.workers)

inputs = [0,1,2,3]

def process(data):
   pl.seed(data)
   output = 0
   for i in pl.arange(1e6):
      this = pl.randn()
      # print('%s: %s' % (i, this))
      output += this
   return output

print('Running parallel...')
sc.tic()
values = [delayed(process)(x) for x in inputs]
results1 = compute(*values, scheduler='distributed')

print(results1)
sc.toc()

print('Running serial...')
sc.tic()
results2 = []
for inputval in inputs:
    results2.append(process(inputval))
print(results2)
sc.toc()


"""
test_datastore.py -- test module for sw_datastore.py
"""

import os
import shutil
import pytest
import sciris as sc
import scirisweb as sw

db_file = 'datastore.db'
db_folder = './temp_test_datastore'

urls = [f'sqlite:///{db_file}', f'file://{db_folder}/']

# Some examples of other URIS
# urls += [
# 'redis://127.0.0.1/3',
# 'sqlite://',
# 'mssql+pyodbc:///?odbc_connect=DRIVER%3D%7BODBC+Driver+13+for+SQL+Server%7D%3BSERVER%3D127.0.0.1%3BDATABASE%testdb%3BUID%3Dtest%3BPWD%3Dtest%3B',
# 'postgresql+psycopg2://test:test@localhost:5432/testdb',
# 'mysql+pymysql://test:test@localhost:3306/testdb?charset=utf8mb4&binary_prefix=true',
# ]

@pytest.mark.parametrize('url', urls)
def test_datastore(url):

    # Reset the database (check flushing works)
    ds = sw.make_datastore(url)

    ds.flushdb()
    ds = sw.make_datastore(url)
    assert len(ds.keys()) == 1 # There should be a datastore settings key present

    # Basic CRUD functionality

    # CREATE
    key_in   = 'testkey'
    data_in  = sc.odict({'foo':[1,2,3], 'bar':[4,5,6]})
    key_out  = ds.saveblob(obj=data_in, key=key_in)

    # READ
    data_out = ds.loadblob(key_in)
    assert key_in  == key_out
    assert data_in == data_out
    assert ds.get('nonexistent') is None
    with pytest.raises(KeyError):
        ds.get('nonexistent', notnone=True)

    # UPDATE
    data_in['foo'][0] = 2
    ds.saveblob(obj=data_in, key=key_in) # This should result in an in-place update
    data_out = ds.loadblob(key_in)
    assert data_out['foo'][0] == 2

    # DELETE
    ds.delete(key_in)
    assert 'foo' not in ds.keys()  # Check it was successfully deleted
    ds.delete('nonexistent')  # If a key doesn't exist, an error should not occur

    # TEST KEY LISTING AND FILTERING
    ds.saveblob(obj='teststr', key='foo')
    ds.saveblob(obj='teststr', key='bar')
    assert {'foo','bar'}.issubset(set(ds.keys()))
    assert set(ds.keys('f*')) == {'foo'}
    assert set(ds.keys('bar')) == {'bar'}

    # TEST EXISTENCE
    assert ds.exists('foo')
    assert not ds.exists('nonexistent')

    # Tidy up
    cleanup = {db_file:os.remove, db_folder:shutil.rmtree}
    for fn,func in cleanup.items():
        try:
            func(fn)
            print('Removed %s' % fn)
        except:
            pass


if __name__ == '__main__':
    for url in urls:
        test_datastore(url)



'''
Usage:

python app.py

then go to localhost:8080 in your browser.
'''

# Imports
import pylab as pl
import sciris as sc
import scirisweb as sw

runserver = True # Choose to run in the frontend or backend

# Create the app
app = sw.ScirisApp(__name__, name="HelloGraph")

# Define the API
@app.route('/showgraph')
def showgraph(n=1000):
    
    # Make graph
    fig = pl.figure()
    ax = fig.add_subplot(111)
    xdata = pl.randn(n)
    ydata = pl.randn(n)
    colors = sc.vectocolor(pl.sqrt(xdata**2+ydata**2))
    ax.scatter(xdata, ydata, c=colors)
    
    # Convert to FE
    graphjson = sw.mpld3ify(fig)  # Convert to dict
    return graphjson  # Return the JSON representation of the Matplotlib figure

# Run the server
if __name__ == "__main__" and runserver:
    app.run()
else:
    showgraph()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 17:15:29 2019

@author: cliffk
"""

from distributed import Worker
from tornado.ioloop import IOLoop
from threading import Thread

loop = IOLoop.current()
t = Thread(target=loop.start, daemon=True)
t.start()

w = Worker('tcp://127.0.0.1:8786', loop=loop)
w.start()  # choose randomly assigned port


#!/usr/bin/env python

'''
Sciris is a flexible open source framework for building scientific web
applications using Python and JavaScript. This library provides the underlying
functions and data structures that support the webapp features, as well as
being generally useful for scientific computing.
'''

from setuptools import setup, find_packages
import os
import sys
import runpy

# Get the current folder
cwd = os.path.abspath(os.path.dirname(__file__))

# Define the requirements and extras
requirements = [
        'sciris', # Basic tools -- NB, this includes numpy, scipy, pandas, and matplotlib as dependencies
        'decorator>=4.1.2', # For API calls
        'redis==2.10.6', # Database -- Redis >=3.0 breaks Celery unfortunately
        'mpld3',    # Rendering plots in the browser
        'werkzeug', # HTTP tools
        'flask>=1.0.0', # Creating the webapp
        'flask-login>=0.4.1', # Handling users
        'flask-session>=0.3.1', # use redis for sessions
        'celery>=4.2', # Task manager
        'twisted>=18.4.0', # Server
        'service_identity', # Identity manager for Celery (not installed with Celery though)
        'pyasn1', # Required for service_identity (but not listed as a dependency!)
        'pyparsing', # Also for processing requests
        'sqlalchemy',
    ],

# Optionally define extras
if 'minimal' in sys.argv:
    print('Performing minimal installation -- celery, twisted, and redis excluded')
    sys.argv.remove('minimal')
    requirements = [
        'sciris', # Basic tools -- NB, this includes numpy, scipy, pandas, and matplotlib as dependencies
        'decorator>=4.1.2', # For API calls
        'mpld3',    # Rendering plots in the browser
        'flask>=1.0.0', # Creating the webapp
        'pyparsing', # Also for processing requests
    ],

# Get version
versionpath = os.path.join(cwd, 'scirisweb', 'sw_version.py')
version = runpy.run_path(versionpath)['__version__']

# Get the documentation
with open(os.path.join(cwd, 'README.md'), "r") as fh:
    long_description = fh.read()

CLASSIFIERS = [
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3.7',
]

setup(
    name='scirisweb',
    version=version,
    author='ScirisOrg',
    author_email='info@sciris.org',
    description='Scientific webapps for Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/sciris/scirisweb',
    keywords=['scientific', 'webapp', 'framework'],
    platforms=['OS Independent'],
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements
    )
